@@ -1,6 +1,6 @@
 // Process server for soft ioc
 // David H. Thompson 8/29/2003
-// Ralph Lange 04/22/2008
+// Ralph Lange 03/23/2010
 // GNU Public License (GPLv3) applies - see www.gnu.org
 
 
@@ -17,6 +17,7 @@
 #include <unistd.h> 
 #include <termios.h>
 #include <sys/ioctl.h>
+#include <sys/select.h>
 #include <string.h>
 
 #include "procServ.h"
@@ -34,6 +35,7 @@ bool   autoRestart = true;       // Enables instant restart of exiting child
 bool   waitForManualStart = false;  // Waits for telnet cmd to manually start child
 bool   shutdownServer = false;   // To keep the server from shutting down
 bool   quiet = false;            // Suppress info output (server)
+bool   setCoreSize = false;      // Set core size for child
 char   *procservName;            // The name of this beast (server)
 char   *childName;               // The name of that beast (child)
 int    connectionNo;             // Total number of connections
@@ -43,9 +45,9 @@ char   toggleRestartChar = 0x14; // Toggle autorestart character (default: ^T)
 char   restartChar = 0x12;       // Restart character (default: ^R)
 char   quitChar = 0x11;          // Quit character (default: ^Q)
 int    killSig = SIGKILL;        // Kill signal (default: SIGKILL)
-rlim_t coreSize = -1;            // Max core size for child
-char   *chDir = NULL;            // Directory to change to before starting child
-char   *myDir = NULL;            // Directory where server was started
+rlim_t coreSize;                 // Max core size for child
+char   *chDir;                   // Directory to change to before starting child
+char   *myDir;                   // Directory where server was started
 time_t holdoffTime = 15;         // Holdoff time between child restarts (in seconds)
 
 pid_t  procservPid;              // PID of server (daemon if not in debug mode)
@@ -73,29 +75,29 @@ void forkAndGo();
 // Checks the command file (existence and access rights)
 int checkCommandFile(const char *command);
 
-void OnSigChild(int);
-int sigChildSet;
-
+// Signal handlers
 void OnSigPipe(int);
+void OnSigTerm(int);
+// Counter used for communication between sig handler and main()
 int sigPipeSet;
 
-struct sigaction sig;
-
 void writePidFile()
 {
-    int pid=getpid();
-    FILE * fp=NULL;
+    int pid = getpid();
+    FILE * fp;
+
+    PRINTF("Writing PID file %s\n", pidFile);
 
     fp = fopen( pidFile, "w" );
     // Don't stop here - just go without
-    if ( fp==NULL ) {
+    if (fp == NULL) {
         fprintf( stderr,
                  "%s: unable to open PID file %s\n",
                  procservName, pidFile );
         return;
     }
-    fprintf( fp, "%d\n", pid );
-    fclose( fp );
+    fprintf(fp, "%d\n", pid);
+    fclose(fp);
 }
 
 char getOptionChar ( const char* buf ) 
@@ -152,23 +154,23 @@ void printVersion()
 
 int main(int argc,char * argv[])
 {
-    time(&procServStart); // What time is it now
-    struct pollfd * pollList=NULL,* ppoll; // Allocate as much space as needed
     int c;
     unsigned int i, j;
     int k;
-    int ctlPort, logPort=0;
+    int ctlPort, logPort = 0;
     char *command;
     bool wrongOption = false;
     char buff[512];
 
+    time(&procServStart);             // remember start time
     procservName = argv[0];
-    myDir = get_current_dir_name();
+    myDir = getcwd(NULL, 512);
     chDir = myDir;
     timeFormat = defaulttimeFormat;
 
     pidFile = getenv( "PROCSERV_PID" );
-    if ( pidFile && ! strcmp( pidFile, "" ) ) pidFile = defaultpidFile;
+    if ( !pidFile || strlen(pidFile) == 0 ) pidFile = defaultpidFile;
+    if ( getenv("PROCSERV_DEBUG") != NULL ) inDebugMode = true;
 
     const int ONE_CHAR_COMMANDS = 2;  // togglerestartcmd, killcmd
 
@@ -217,7 +219,10 @@ int main(int argc,char * argv[])
 
         case 'C':                                 // Core size
             k = atoi( optarg );
-            if ( k >= 0 ) coreSize = k;
+            if ( k >= 0 ) {
+                coreSize = k;
+                setCoreSize = true;
+            }
             break;
 
         case 'c':                                 // Dir to change to
@@ -340,14 +345,15 @@ int main(int argc,char * argv[])
     }
 
     // Single command characters should be ignored, too
-    if ( ignChars == NULL && ( killChar || toggleRestartChar ) )
-        ignChars = (char*) calloc( 1 + ONE_CHAR_COMMANDS, 1);
-    if ( killChar )
-        strcat ( ignChars, &killChar );
-    if ( toggleRestartChar )
-        strcat ( ignChars, &toggleRestartChar );
+    if (ignChars == NULL && (killChar || toggleRestartChar))
+        ignChars = (char*) calloc(1 + ONE_CHAR_COMMANDS, 1);
+    if (killChar)
+        strncat (ignChars, &killChar, 1);
+    if (toggleRestartChar)
+        strncat (ignChars, &toggleRestartChar, 1);
 
-    // set up available server commands message
+    // Set up available server commands message
+    PRINTF("Setting up messages\n");
     sprintf( infoMessage3, "@@@ Use %s%c to restart the child now, %s%c to quit the server" NL,
              restartChar < 32 ? "^" : "",
              restartChar < 32 ? restartChar + 64 : restartChar,
@@ -368,12 +374,17 @@ int main(int argc,char * argv[])
 
     if ( checkCommandFile( command ) ) exit( errno );
 
-    sig.sa_handler=&OnSigChild;
-    sigaction(SIGCHLD,&sig,NULL);
-    sig.sa_handler=&OnSigPipe;
-    sigaction(SIGPIPE,&sig,NULL);
+    struct sigaction sig;
+    memset(&sig, 0, sizeof(sig));
+
+    PRINTF("Installing signal handlers\n");
+    sig.sa_handler = &OnSigPipe;              // sigaction() needed for Solaris
+    sigaction(SIGPIPE, &sig, NULL);
+    sig.sa_handler = &OnSigTerm;
+    sigaction(SIGTERM, &sig, NULL);
 
     // Make an accept item to listen for control connections
+    PRINTF("Creating control listener\n");
     try
     {
 	connectionItem *acceptItem = acceptFactory( ctlPort, ctlPortLocal );
@@ -389,6 +400,7 @@ int main(int argc,char * argv[])
 
     if ( logPort ) {
         // Make an accept item to listen for log connections
+        PRINTF("Creating log listener\n");
         try
         {
             connectionItem *acceptItem = acceptFactory( logPort, logPortLocal, true );
@@ -405,8 +417,6 @@ int main(int argc,char * argv[])
 
     procservPid=getpid();
 
-    if ( getenv("PROCSERV_DEBUG") != NULL ) inDebugMode = true;
-
     if ( inDebugMode == false )
     {
 	forkAndGo();
@@ -414,7 +424,7 @@ int main(int argc,char * argv[])
     }
     else
     {
-	debugFD = 1;          // Enable debug messages
+        debugFD = 1;          // Enable debug messages
     }
 
     // Open log file
@@ -430,10 +440,10 @@ int main(int argc,char * argv[])
     }
 
     // Record some useful data for managers 
-    sprintf( infoMessage1, "@@@ procServ server PID: %d" NL
+    sprintf( infoMessage1, "@@@ procServ server PID: %ld" NL
              "@@@ Server startup directory: %s" NL
              "@@@ Child startup directory: %s" NL,
-             getpid(),
+             (long) getpid(),
              myDir,
              chDir);
     if ( strcmp( childName, command ) )
@@ -449,72 +459,59 @@ int main(int argc,char * argv[])
     // Run here until something makes it die
     while ( ! shutdownServer )
     {
-	char buf[100];
-	int nPoll = 0;      // local copy of the # items to poll
-	int nPollAlloc = 0; // How big is pollList right now
-	int pollStatus;     // What poll returns
+        char buf[100];
+        connectionItem * p;
+        fd_set fdset;              // FD stuff for select()
+        int fd, nFd;
+        int ready;                 // select() return value
+        struct timeval timeout;
 
-	connectionItem * p ;
+        if (sigPipeSet > 0) {
+            sprintf( buf, "@@@ Got a sigPipe signal: Did the child close its tty?" NL);
+            SendToAll( buf, strlen(buf), NULL );
+            sigPipeSet--;
+        }
 
-	if (sigPipeSet>0)
-	{
-	    PRINTF("Got a sigPipe\n");
-	    sprintf( buf, "@@@ Got a sigPipe signal: Did the IOC close its tty?" NL);
-	    SendToAll( buf, strlen(buf), NULL );
-	    sigPipeSet--;
-	}
-	// Adjust the poll data array
-	if (nPollAlloc != connectionNo)
-	{
-	    if (pollList) free(pollList);
-	    pollList = (pollfd*) malloc( connectionNo * sizeof(struct pollfd) );
-	    nPollAlloc = connectionNo;
-	}
+        // Prepare FD set for select()
+        p = connectionItem::head;
+        nFd = -1;
+        FD_ZERO(&fdset);
+        while (p) {
+            if ((fd = p->getFd()) > -1) {     // Connection needs to be watched
+                if (fd > nFd) nFd = fd;
+                FD_SET(fd, &fdset);
+            }
+            p = p->next;
+        }
+        nFd++;
+        timeout.tv_sec = 0;                   // select() timeout: 0.5 sec
+        timeout.tv_usec = 500000;
 
-	// Load the socket number and flags
-	ppoll=pollList;
-	nPoll=0;
-	p = connectionItem::head;
-	while(p)
-	{
-	    if (p->SetPoll(ppoll))
-	    {
-	    	ppoll++; // Move to the nexr fd slot
-		nPoll++; // How many to poll
-	    }
-	    p=p->next;
-	}
-	
-	// Now do the poll to find new data
-	pollStatus=poll(pollList,nPoll,500);
+        ready = select(nFd, &fdset, NULL, NULL, &timeout);
 
-	// handle what poll returns
-        if (pollStatus==0)
-	{
-	    // Go clean up dead connections
-	    OnPollTimeout();
-	    connectionItem * npi; 
+        if (ready == -1) {                    // Error
+            perror("Error in select() call");
+        } else if (ready == 0) {              // Timeout
+            // Go clean up dead connections
+            OnPollTimeout();
+            connectionItem * npi; 
 
-	    // Pick up the process item if it dies
-	    // This call returns NULL if the process item lives
-	    if (processFactoryNeedsRestart())
-	    {
-	    	npi= processFactory(argc-optind-2,argv+optind+1);
-	    	if (npi) AddConnection(npi);
-	    }
-	}
-	else
-	{
-	    // Loop thru all of the connectionItems until we find the one(s) that need to
-	    // do I/O
-	    p = connectionItem::head;
-	    nPoll=0;
-	    while(p && nPoll<pollStatus)
-	    {
-		if (p->OnPoll()) nPoll++;
-		p=p->next;
-	    }
-	}
+            // Pick up the process item if it dies
+            // This call returns NULL if the process item lives
+            if (processFactoryNeedsRestart())
+            {
+                npi= processFactory(argc-optind-2,argv+optind+1);
+                if (npi) AddConnection(npi);
+            }
+        } else {                              // Work to be done
+            // Loop through all connections
+            p = connectionItem::head;
+            while (p) {
+                if (FD_ISSET(p->getFd(), &fdset)) p->readFromFd();
+                p = p->next;
+            }
+            OnPollTimeout();
+        }
     }
 }
 
@@ -535,7 +532,7 @@ void SendToAll(const char * message,int count,const connectionItem * sender)
     }
 
     while ( p ) {
-	if ( p->isProcess() ) 
+	if ( p->isProcess() )
 	{
 	    // Non-null senders that are not processes can send to processes
 	    if (sender && !sender->isProcess()) p->Send(message,count);
@@ -555,21 +552,20 @@ void OnPollTimeout()
 {
     pid_t pid;
     int wstatus;
-    connectionItem * p = connectionItem::head;
+    connectionItem *pc, *pn;
     char buf[128];
-    
-    if (sigChildSet)
-    {
-	pid = wait(&wstatus);
-	while(p)
+
+    pid = waitpid(-1, &wstatus, WNOHANG);
+    if (pid > 0 ) {
+        pc = connectionItem::head;
+        while(pc)
 	{
-	    p->OnWait(pid);
-	    p=p->next;
+            pc->markDeadIfChildIs(pid);
+            pc=pc->next;
 	}
-	sigChildSet--;
 
 	sprintf( buf, NL "@@@ @@@ @@@ @@@ @@@" NL
-                 "@@@ Received a sigChild for process %d." , pid );
+                 "@@@ Received a sigChild for process %ld." , (long) pid );
 
 	if (WIFEXITED(wstatus))
 	{
@@ -586,26 +582,21 @@ void OnPollTimeout()
 	SendToAll( buf, strlen(buf), NULL );
     }
 
-    p = connectionItem::head;
-    while (p)
+    // Clean up connections
+    pc = connectionItem::head;
+    while (pc)
     {
-    
-	if (p->IsDead())
-	{
-	    DeleteConnection(p);
-	    // p is now invalid. 
-	    // break the loop now and we will
-	    // get another chance later.
-	    break;
-	}
-	p=p->next;
+        pn = pc->next;
+        if (pc->IsDead()) DeleteConnection(pc);
+        pc = pn;
     }
 }
 
 // Call this to add the item to the list of connections
 void AddConnection(connectionItem * ci)
 {
-	if (connectionItem::head )
+    PRINTF("Adding connection %p to list\n", ci);
+    if (connectionItem::head )
 	{
 	    ci->next=connectionItem::head;
 	    ci->next->prev=ci;
@@ -615,13 +606,13 @@ void AddConnection(connectionItem * ci)
 	ci->prev=NULL;
 	connectionItem::head=ci;
 	connectionNo++;
-	PRINTF("Adding connection\n");
 }
 
 
 void DeleteConnection(connectionItem *ci)
 {
-	if (ci->prev) // Not the head
+    PRINTF("Deleting connection %p\n", ci);
+    if (ci->prev) // Not the head
 	{
 		ci->prev->next=ci->next;
 	}
@@ -630,68 +621,69 @@ void DeleteConnection(connectionItem *ci)
 		connectionItem::head = ci->next;
 	}
 	if (ci->next) ci->next->prev=ci->prev;
-	delete ci;
+        delete ci;
 	connectionNo--;
 	assert(connectionNo>=0);
-	PRINTF("Deleting connection\n");
 }
 
-void OnSigChild(int)
-{
-	sigChildSet++;
-}
 void OnSigPipe(int)
 {
-	sigPipeSet++;
+    PRINTF("SigPipe received\n");
+    sigPipeSet++;
+}
+
+void OnSigTerm(int)
+{
+    PRINTF("SigTerm received\n");
+    processFactorySendSignal(killSig);
+    shutdownServer = true;
 }
 
 // Fork the daemon and exit the parent
 void forkAndGo()
 {
-    pid_t p=fork();
-    char buf[] = "/dev/null";
+    pid_t p;
     int fh;
     struct stat statBuf;
 
-    fstat( 1, &statBuf ); // Find out what stdout is
+    fstat(1, &statBuf);      // Find out what stdout is
 
-    if ( p ) // I am the parent
-    {
-        if ( !quiet ) {
-            fprintf( stderr, "%s: spawning daemon process: %d\n", procservName, p );
-            if ( logFile == NULL ) {
-                if ( S_ISREG(statBuf.st_mode) )
-                    fprintf( stderr, "The open file on stdout will be used as a log file.\n" );
+    if ((p = fork()) < 0) {  // Fork failed
+        perror("Could not fork daemon process");
+        exit(errno);
+    } else if (p > 0) {      // I am the parent (foreground command)
+        if (!quiet) {
+            fprintf(stderr, "%s: spawning daemon process: %ld\n", procservName, (long) p);
+            if (logFileFD == -1) {
+                if (S_ISREG(statBuf.st_mode))
+                    fprintf(stderr, "The open file on stdout will be used as a log file.\n");
                 else
-                    fprintf( stderr, "No log file specified and stdout is not a file "
-                             "- no log will be kept.\n" );
+                    fprintf(stderr, "No log file specified and stdout is not a file "
+                            "- no log will be kept.\n" );
             }
         }
-	exit(0);
+        exit(0);
+    } else {                 // I am the child (background daemon)
+        procservPid = getpid();
+
+        // Use stdout as log file, if no file is set and stdout is a file
+        if (logFileFD == -1 && S_ISREG(statBuf.st_mode)) logFileFD = dup(1);
+
+        // Redirect stdin, stdout, stderr to /dev/null
+        char buf[] = "/dev/null";
+        fh = open(buf, O_RDWR);
+        if (fh < 0) { perror(buf); exit(-1); }
+        close(0); close(1); close(2);
+        dup(fh); dup(fh); dup(fh);
+        close(fh);
+
+        // Make sure we are not attached to a terminal
+        fh = open("/dev/tty", O_RDWR);
+        if (fh >= 0) {                // It's a terminal
+            ioctl(fh, TIOCNOTTY);     // Detach from /dev/tty
+            close(fh);
+        }
     }
-    procservPid = getpid();
-
-    // p==0
-    // The daemon starts up here
-
-    // Redirect all of the I/O away from /dev/tty
-    fh = open( buf, O_RDWR );
-    if ( fh<0 ) { perror( buf ); exit( -1 ); }
-
-    if ( logFileFD == -1 && S_ISREG(statBuf.st_mode) )
-    {
-	logFileFD = dup(1);
-    }
-    close(0); close(1); close(2);
-    
-    dup(fh); dup(fh); dup(fh);
-    close(fh);
-
-    // Now, make sure we are not attached to a terminal
-    fh = open( "/dev/tty", O_RDWR );
-    if ( fh<0 ) return;         // Not a terminal
-    ioctl( fh, TIOCNOTTY );     // Detatch from /dev/tty
-    close( fh );
 }
 
 
@@ -700,31 +692,28 @@ void forkAndGo()
 int checkCommandFile(const char * command)
 {
     struct stat s;
-    int ngroups_max=1+sysconf(_SC_NGROUPS_MAX);
-    gid_t * groups=new gid_t[ngroups_max];
-    ngroups_max=getgroups(ngroups_max,groups);
     mode_t min_permissions=S_IRUSR|S_IXUSR|S_IXGRP|S_IRGRP|S_IROTH|S_IXOTH;
 
-    // chdir if possible (to allow relative command)
-    if ( chDir && chdir( chDir ) ) perror( chDir );
+    PRINTF("Checking command file validity\n");
 
-    if (stat(command,&s))
-    {
-	perror(command);
-	return -1;
+    // chdir if possible (to allow relative command)
+    if (chDir && chdir(chDir)) perror(chDir);
+
+    if (stat(command,&s)) {
+        perror(command);
+        return -1;
     }
 
     // Back to where we came from
-    if ( chdir( myDir ) ) perror( myDir );
+    if (chdir(myDir)) perror(myDir);
 
-    if (!S_ISREG(s.st_mode))
-    {
-	fprintf(stderr,"%s: %s is not a regular file\n", procservName, command);
-	return -1;
+    if (!S_ISREG(s.st_mode)) {
+        fprintf(stderr, "%s: %s is not a regular file\n", procservName, command);
+        return -1;
     }
-    if ( min_permissions == (s.st_mode & min_permissions) ) return 0; // This is great!
+    if (min_permissions == (s.st_mode & min_permissions)) return 0; // This is great!
     // else
-    fprintf( stderr, "%s: Warning - Please change permissions on %s to at least -r-xr-xr-x\n"
+    fprintf(stderr, "%s: Warning - Please change permissions on %s to at least -r-xr-xr-x\n"
              "procServ may not be able to continue without execute permission!\n",
              procservName, command);
     return 0;
